/* file command.c */
/*
 *      narcisse81 - a graphical frontend for diogene87 
 *	Copyright (C) 2005 - Dominique Parisot
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <gtk/gtk.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include "command.h"
#include "global.h"

#define TERMINATOR "\r\n"
#define PROMPT "Diogene87>"

static int fd = -1;
static gboolean busy = FALSE;

static gchar* socket_read(void);
static gboolean socket_write(char* buffer);
/*****************************************************************************/
void nc81_command_close_connection(void)
{
	if (fd == -1)
		return;
      	shutdown (fd, 2);
	close(fd);
	fd = -1;
}

/*****************************************************************************/
static gboolean socket_open(void)
{
	gchar *character;
	gchar* buffer;
	int ip_port;
	struct sockaddr_in address;
	struct hostent *hostnameent;
	int length, result;
	ip_port = atoi(nc81_config_ip_port);
	/* nothing to do if connection is already opened */
	if (fd != -1)
		return TRUE;
	/* convert hostname to IP address */
	hostnameent = gethostbyname (nc81_config_server);
	if (!hostnameent)
	{
		g_print("command : erreur gethostbyname\n");
		return FALSE;
	}
	/* create the socket */
	if ((fd = socket (AF_INET, SOCK_STREAM, 0)) == -1)
	{
		g_print("command : erreur socket\n");
		return FALSE;
	}
	/* connect to the remote */
	address.sin_family = AF_INET;
	address.sin_addr = *(struct in_addr*) *hostnameent->h_addr_list;
	address.sin_port = htons (ip_port);
	length = sizeof (address);
	result = connect (fd, (struct sockaddr *) &address, length);
	if (result == -1)
	{
		g_print("command : erreur connect\n");
		nc81_command_close_connection(); 
		return FALSE;
	}
	/* read and verify the daemon banner */
	if ((buffer = socket_read()) == NULL)
	{
		nc81_command_close_connection(); 
		return FALSE;
	}
	if (strncmp(buffer,"diogene87", 9) != 0){
		g_print("wrong banner '%s' !\n", buffer);
		g_free(buffer);
		nc81_command_close_connection();
		return FALSE;
	}
	g_free(buffer);

	/* send the identification */
	buffer = g_strdup_printf("user \"%s\" \"%s\"%s", 
			nc81_config_user, 
			nc81_config_password, 
			TERMINATOR);
	if (socket_write(buffer) == FALSE)
	{
		/* TODO free memory */
		g_free(buffer);
		nc81_command_close_connection();
		return FALSE;
	}
	g_free(buffer);

	/* read the result of the identification */
	if ((buffer = socket_read()) == NULL)
	{
		/* TODO free memory */
		g_print("incorrect identification !\n");
		nc81_command_close_connection();
		return FALSE;
	}
	/* remove the last line_feed */
	if ((character = strrchr(buffer, '\r')) != NULL)
		*character = '\0';
	/* check that identification is good */
	if (strcmp(buffer, "OK") != 0)
	{
		/* TODO free memory */
		g_free(buffer);
		nc81_command_close_connection();
		return FALSE;
	}
	g_free(buffer);

	return TRUE;
}
/*****************************************************************************/
static gchar* socket_read(void)
{
	fd_set readfds;
	struct timeval tv;
	int result;
	gchar* buffer;
	int nread, ntoread;

	/* wait buffer_str (maximum 5 seconds) */
	tv.tv_sec = 5;
	tv.tv_usec = 0;
	FD_ZERO(&readfds);
	FD_SET(fd, &readfds);
	result = select(fd+1, &readfds, NULL, NULL, &tv);
	if (result <= 0)
		return NULL;

	/* get the number of bytes to read */
	if (ioctl (fd, FIONREAD, &ntoread) == -1)
		return NULL;
	if (ntoread == 0)
		return NULL;

	/* read the buffer */
	if ((buffer = g_malloc(ntoread + 1)) == NULL)
		return NULL;
	memset(buffer, 0, ntoread + 1);
	nread = read (fd, buffer, ntoread);
	if (nread != ntoread)
	{
		g_free(buffer);
		return NULL;
	}
	return buffer;
}
/*****************************************************************************/
static gboolean socket_write(char* buffer)
{
	int nwrite;
	int buffer_size = strlen(buffer);
	nwrite = write (fd, buffer, buffer_size);
	if (nwrite < 0)
		return FALSE;
	return TRUE;
}
/*****************************************************************************/
void nc81_command_send(gchar* command, 
		void (*callback_function) (gchar** fields, void* data),
		void* data, 
		gchar* valuex, gchar* valuey, gchar* valuez)
{
	gchar* previous_line;
	gchar *error_msg;
	gchar *buffer;
	gchar **fields;
	gchar *character;
	GString *buffer_str;
	gint line_length;
	gboolean terminated;

	if (busy == TRUE)
	{
		g_print("process busy, can't execute [%s]\n", command);
		return;
	}
	busy = TRUE;

	/* replace %? of the command by variable value */
	buffer_str = g_string_new("");
	for (character = command; *character!='\0'; character++)
	{
		if (*character=='%')
		{
			switch (*(++character))
			{
				case 'r': /* referential name */
					assert(nc81_ref_name != NULL);
					g_string_append(buffer_str, 
							nc81_ref_name);
					break;
				case 't': /* task name */
					assert(nc81_task_name != NULL);
					g_string_append(buffer_str, 
							nc81_task_name);
					break;
				case 'j': /* job name */
					assert(nc81_job_name != NULL);
					g_string_append(buffer_str, 
							nc81_job_name);
					break;
				case 'd': /* day name */
					assert(nc81_day_name != NULL);
					g_string_append(buffer_str, 
							nc81_day_name);
					break;
				case 'T': /* target name */
					assert(nc81_target_name != NULL);
					g_string_append(buffer_str, 
							nc81_target_name);
					break;
				case 'Q': /* queue name */
					assert(nc81_queue_name != NULL);
					g_string_append(buffer_str, 
							nc81_queue_name);
					break;
				case 'R': /* rule name */
					assert(nc81_rule_name != NULL);
					g_string_append(buffer_str, 
							nc81_rule_name);
					break;
				case 'u': /* user name */
					assert(nc81_user_name != NULL);
					g_string_append(buffer_str, 
							nc81_user_name);
					break;
				case 'v': /* variable name */
					assert(nc81_variable_name != NULL);
					g_string_append(buffer_str, 
							nc81_variable_name);
					break;
				case 'x': /* variable value */
					g_string_append(buffer_str, 
						(valuex==NULL)?"":valuex);
					break;
				case 'y': /* variable value */
					g_string_append(buffer_str, 
						(valuey==NULL)?"":valuey);
					break;
				case 'z': /* variable value */
					g_string_append(buffer_str, 
						(valuez==NULL)?"":valuez);
					break;
			}
		}
		else
		{
			g_string_append_c(buffer_str, *character);
		}
	}
	g_print("*command [%s]\n", buffer_str->str);

	/* open/reopen connection */
	if (socket_open() == FALSE)
	{
		nc81_message_ok("Connection error !");
		g_print("connection error !\n");
		g_string_free(buffer_str, TRUE);
		busy = FALSE;
		return;
	}

	/* send the command */
	buffer = g_strdup_printf("%s%s", buffer_str->str, TERMINATOR);
	g_string_free(buffer_str, TRUE);
	if (socket_write(buffer) == FALSE)
	{
		nc81_message_ok("Can't send command !");
		/* TODO free memory */
		nc81_command_close_connection();
		busy = FALSE;
		return;
	}
	g_free(buffer);

	/* read the result of the user command */
	previous_line = NULL;
	buffer_str = g_string_new("");
	terminated = FALSE;
	while (!terminated)
	{
		/* read the result (or a part of the result) */
		if ((buffer = socket_read()) == NULL)
		{
			/* TODO free memory */
			nc81_message_ok("Read error !");
			g_print("Can't read socket !\n");
			nc81_command_close_connection();
			g_string_free(buffer_str, TRUE);
			busy = FALSE;
			return;
		}
		g_string_append(buffer_str, buffer);
		g_free(buffer);

		/* split the buffer into lines */
		while((character = strstr(buffer_str->str, TERMINATOR)) != NULL)
		{
			line_length = character - buffer_str->str;
			buffer_str->str[line_length] = '\0';
			buffer_str->str[line_length+1] = '\0';

			/* prompt */
			if (strcmp(character + 2, PROMPT) == 0)
			{
				terminated = TRUE;
				break;
			}

			/* extract the line from the buffer */
			buffer_str->str[line_length] = '\t';
			buffer_str->str[line_length+1] = '\0';
			fields = g_strsplit(buffer_str->str, "\t", 0);
			callback_function(fields, data);
			g_strfreev(fields);

			/* remove the line from the buffer */
			if (previous_line != NULL)
				g_free(previous_line);
			previous_line = g_strdup(buffer_str->str);
			g_string_erase(buffer_str, 0, line_length + 2);


		} /* next line */
	} /* next buffer_str buffer */

	/* interpret the exit status */
	if (strcmp(buffer_str->str, "OK") != 0)
	{
		error_msg = g_strdup_printf("Error : %s", 
				(previous_line == NULL)?"?":previous_line);
		nc81_message_ok(error_msg);
		g_free(error_msg);
	}

	if (previous_line != NULL)
		g_free(previous_line);
	g_string_free(buffer_str, TRUE);
	busy = FALSE;
	/* connection is not closed : it will be used for the next command */
}
/*****************************************************************************/
