/* file graph.c */
/*
 *      narcisse81 - a graphical frontend for diogene87 
 *	Copyright (C) 2005 - Dominique Parisot
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA
 *
 */


#include <assert.h>
#include <string.h>
#include <gtk/gtk.h>
#include "global.h"
#include "command.h"

typedef enum 
{
	nc81_graph_view_ref_task,
	nc81_graph_view_ref_job,
	nc81_graph_view_exp_task,
	nc81_graph_view_exp_job
}
nc81_graph_view_type;

static nc81_graph_view_type view;

/****************************************************************************/
typedef struct
{
	gchar* name;
	gchar* description;
	gint x, y;
	GSList *upstream_link;
	GSList *downstream_link;
} nc81_graph_label;

typedef enum
{
	nc81_graph_link_alway,
	nc81_graph_link_on_success,
	nc81_graph_link_on_error
}
nc81_graph_link_type;

typedef struct
{
	nc81_graph_label *upstream;	
	nc81_graph_label *downstream;	
	nc81_graph_link_type type;
} nc81_graph_link;

static GSList *label_list = NULL;
static GSList *link_list= NULL;
static gint label_width = 35;
static gint label_height = 12;
static gint delta_width = 3;
static gint delta_height = 12;

static gint posx, posy;
static nc81_graph_label *selected_label = NULL;

static guint msg_id = 0;
static GtkWidget *statusbar;

/****************************************************************************/
static void list_label_fill_callback(gchar** fields, void* data)
{
	/* rtl, rjl, xtl, xjl */
	/* -> task/job - description */
	nc81_graph_label *label;
	label = g_malloc(sizeof(nc81_graph_label));
	assert(label != NULL);
	label->name = g_strdup(fields[0]);
	assert(label->name != NULL);
	label->description = g_strdup(fields[1]);
	assert(label->description != NULL);

	label_list = g_slist_append(label_list, label);
	label->upstream_link = NULL;
	label->downstream_link = NULL;
	label->x = -1;
	label->y = -1;
	/*
	g_print("label : %s - %s\n", label->name, label->description);
	*/
}
/****************************************************************************/
static nc81_graph_label*
nc81_graph_label_find(gchar* name)
{
	GSList *label_item;
	nc81_graph_label *label;

	label_item = label_list;
	while (label_item != NULL)
	{
		label = label_item->data;
		if (strcmp(name, label->name) == 0)
			break;
		label_item = g_slist_next(label_item);
	}
	return (label_item == NULL) ? NULL : label;
}
/****************************************************************************/
static void list_link_fill_callback(gchar** fields, void* data)
{
	nc81_graph_link *link;
	link = g_malloc(sizeof(nc81_graph_link));
	assert(link != NULL);

	/* find upstream/downstream label */
	link->upstream = nc81_graph_label_find(fields[0]);
	assert(link->upstream != NULL);
	link->downstream = nc81_graph_label_find(fields[1]);
	assert(link->downstream != NULL);

	if (strcmp(fields[2], "always") == 0)
		link->type = nc81_graph_link_alway;
	if (strcmp(fields[2], "on_success") == 0)
		link->type = nc81_graph_link_on_success;
	if (strcmp(fields[2], "on_error") == 0)
		link->type = nc81_graph_link_on_error;

	link_list = g_slist_append(link_list, link);
	link->upstream->downstream_link =
		g_slist_append(link->upstream->downstream_link, link);
	link->downstream->upstream_link =
		g_slist_append(link->downstream->upstream_link, link);

	/*
	g_print("link : %s - %s (%i)\n", 
			link->upstream->name, 
			link->downstream->name,
			link->type);
			*/

}
/****************************************************************************/
void nc81_graph_label_follow(nc81_graph_label* label)
{
	GSList* label_item;
	nc81_graph_link* link;
	if (label == NULL)
		return;

	/* if we are at top level, ignore task which are not at the top level */
	if (posy == 0 && label->upstream_link != NULL)
		return;

	/*
	if (label->x != -1)
		posx++;
		*/

	/* if label is already positionned, nothing to do */
	if (label->x != -1 && label->y >= posy)
		return;

	/*
	g_print("<follow %s %i> ", label->name, posx);
	*/

	if (label->x == -1)
	{
		label->x = posx;
		if (label->downstream_link == NULL)
			posx++;
	}
	if (label->y < posy)
		label->y = posy;

	/* follow each downstream link */
	posy ++;
	for (label_item = label->downstream_link; label_item != NULL;
			label_item = g_slist_next(label_item))
	{
		link = label_item->data;
		assert(link != NULL);
		nc81_graph_label_follow(link->downstream);
	}
	posy --;
	return;
}
/****************************************************************************/
gboolean nc81_graph_button_press(GtkWidget* drawing_area, GdkEventButton *ev, gpointer data)
{
	GSList* list_item;
	nc81_graph_label* label;
	char* msg;
	gint x, y;
	x = ev->x;
	y = ev->y;
	
	for (list_item = label_list; list_item != NULL; 
		list_item = g_slist_next(list_item))
	{
		label = list_item->data;
		if (label->x <= x && x <= label->x+label_width  
				&& label->y <= y && y <= label->y+label_height)
		{
			/* show description in statusbar */
			gtk_statusbar_pop(GTK_STATUSBAR(statusbar), msg_id);
			msg = g_strdup_printf("%s : %s",
					label->name, label->description);
			gtk_statusbar_push(GTK_STATUSBAR(statusbar), 
					msg_id, msg);
			g_free(msg);
			if (selected_label != label)
			{
				selected_label = label;
				gtk_widget_queue_draw(GTK_WIDGET(drawing_area));
			}
			break;
		}
	}

	return TRUE;
}
/****************************************************************************/
gboolean nc81_graph_redraw(GtkWidget* drawing_area, GdkEventExpose *ev)
{
	GSList* list_item;
	GdkColor white_color;
	GdkColor blue_color;
	GdkColor black_color;
	GdkColor yellow_color;
	gint width, height;
	GdkColormap *colormap;
	GdkWindow *window;
	GdkGC *gc;
	GdkFont *font;
	gint x,y, x1,y1, x2,y2;

	nc81_graph_label* label;
	nc81_graph_link* link;

	if (GTK_WIDGET_DRAWABLE(drawing_area) == FALSE)
		return FALSE;

	window = drawing_area->window;

	colormap = gdk_window_get_colormap(window);
	gdk_window_get_size(window, &width, &height);

	if (gdk_color_parse("white", &white_color) == TRUE)
	gdk_colormap_alloc_color(colormap, &white_color, FALSE, TRUE);
	gdk_color_parse("black", &black_color);
	gdk_colormap_alloc_color(colormap, &black_color, FALSE, TRUE);
	gdk_color_parse("yellow", &yellow_color);
	gdk_colormap_alloc_color(colormap, &yellow_color, FALSE, TRUE);
	gdk_color_parse("red", &blue_color);
	gdk_colormap_alloc_color(colormap, &blue_color, FALSE, TRUE);

	gdk_window_set_background(window, &white_color);

	gc = gdk_gc_new(window);
	gdk_gc_set_clip_rectangle(gc, &(ev->area));

	/*
	font = gdk_font_load("-*-fixed-*-*-*-*-*-120-*-*-*-*-*-*");
	*/
	font = gdk_font_load("-*-fixed-*-*-*-*-*-60-*-*-*-*-*-*");
	gdk_gc_set_font(gc, font);

	/* clear background */
	gdk_gc_set_foreground(gc, &white_color);
	gdk_draw_rectangle(window, gc, TRUE, 0, 0, width, height);

	/* draw each label */
	for (list_item = label_list; list_item != NULL; 
		list_item = g_slist_next(list_item))
	{
		label = list_item->data;
		x = label->x;
		y = label->y;
		gdk_gc_set_foreground(gc, &yellow_color);
		gdk_draw_rectangle(window, gc, TRUE, x, y, 
				label_width, label_height);
		gdk_gc_set_foreground(gc, &black_color);
		gdk_draw_rectangle(window, gc, FALSE, x, y, 
				label_width, label_height);
		gdk_draw_string(window, font, gc, 
				x + 1, y + delta_height - 3,
				label->name);
	}

	/* draw each link */
	for (list_item = link_list; list_item != NULL; 
		list_item = g_slist_next(list_item))
	{
		link = list_item->data;
		/* upstream */
		x1 = link->upstream->x + label_width / 2;
		y1 = link->upstream->y + label_height;
		/* downstream */
		x2 = link->downstream->x + label_width / 2;
		y2 = link->downstream->y;
		/* middle */
		y = y2 - delta_height / 2;
		gdk_draw_line(window, gc, x1, y1, x1, y);
		gdk_draw_line(window, gc, x1, y, x2, y);
		gdk_draw_line(window, gc, x2, y, x2, y2);
	}
	/* redraw the selected label and its links */
	if (selected_label != NULL)
	{
		label = selected_label;
		gdk_gc_set_line_attributes(gc, 2,
				GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER);
		gdk_gc_set_foreground(gc, &blue_color);
		gdk_draw_rectangle(window, gc, FALSE, label->x, label->y, 
				label_width, label_height);

		for (list_item = link_list; list_item != NULL; 
			list_item = g_slist_next(list_item))
		{
			link = list_item->data;
			if (link->upstream != selected_label 
					&& link->downstream != selected_label)
				continue;
			/* upstream */
			x1 = link->upstream->x + label_width / 2;
			y1 = link->upstream->y + label_height;
			/* downstream */
			x2 = link->downstream->x + label_width / 2;
			y2 = link->downstream->y;
			/* middle */
			y = (y1 + y2) / 2;
			y = y2 - delta_height / 2;
			gdk_draw_line(window, gc, x1, y1, x1, y);
			gdk_draw_line(window, gc, x1, y, x2, y);
			gdk_draw_line(window, gc, x2, y, x2, y2);
		}
	}
	gdk_gc_destroy(gc);
	return TRUE;
}
/****************************************************************************/
static void nc81_graph_destroy_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	GSList* list_item;
	nc81_graph_label *label;
	nc81_graph_link *link;
	/* TODO free memory : label_list et linklist */
	g_print("free memory\n");
	for (list_item = label_list; list_item != NULL; 
		list_item = g_slist_next(list_item))
	{
		label = list_item->data;
		g_free(label->name);
		g_free(label->description);
		g_free(label);
	}
	g_slist_free(label_list);
	for (list_item = link_list; list_item != NULL; 
		list_item = g_slist_next(list_item))
	{
		link = list_item->data;
		g_free(link);
	}
	g_slist_free(link_list);
}
/****************************************************************************/
static GtkWidget* nc81_graph_new(gchar* cmd_list_label, gchar* cmd_list_link)
{
	GtkWidget *vbox;
	GtkWidget *drawing_area;
	GtkWidget *scrolled;
	GtkWidget *eventbox;
	GSList* label_item;
	nc81_graph_label* label;
	gint max_x = 0; 
	gint max_y = 0;
	/* TODO clear memory */
	label_list = NULL;
	link_list = NULL;

	/* get task/job list */
	nc81_command_send(cmd_list_label, 
			list_label_fill_callback, NULL, NULL, NULL, NULL);
	/* get dependence list */
	nc81_command_send(cmd_list_link, 
			list_link_fill_callback, NULL, NULL, NULL, NULL);

	/* follow each top label */
	posx = 0;
	for (label_item = label_list; label_item != NULL; 
		label_item = g_slist_next(label_item))
	{
		label = label_item->data;
		posy = 0;
		nc81_graph_label_follow(label);
	}
	
	/* find min and max */
	for (label_item = label_list; label_item != NULL; 
		label_item = g_slist_next(label_item))
	{
		label = label_item->data;
		if (label->x > max_x) 
			max_x = label->x;
		if (label->y > max_y) 
			max_y = label->y;
		label->x = label->x * (label_width + delta_width);
		label->y = label->y * (label_height + delta_height);
	}
	g_print("max x:%i y:%i\n", max_x, max_y);
	
	/* create drawing area */
	vbox = gtk_vbox_new(FALSE, 0);
	drawing_area = gtk_drawing_area_new();
	gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area), 
			(max_x +1) * (label_width + delta_width),
			(max_y +1) * (label_height + delta_height));
	gtk_signal_connect(GTK_OBJECT(drawing_area), "expose_event",
			GTK_SIGNAL_FUNC(nc81_graph_redraw), NULL);

	scrolled = gtk_scrolled_window_new(NULL,NULL);
	gtk_box_pack_start(GTK_BOX(vbox), scrolled, TRUE, TRUE,0);

	gtk_scrolled_window_set_policy(
			GTK_SCROLLED_WINDOW(scrolled),
			GTK_POLICY_AUTOMATIC,
			GTK_POLICY_AUTOMATIC);

	eventbox = gtk_event_box_new();
	gtk_signal_connect(GTK_OBJECT(eventbox), "button_press_event",
			GTK_SIGNAL_FUNC(nc81_graph_button_press), NULL);
	gtk_container_add(GTK_CONTAINER(eventbox), drawing_area);

	gtk_scrolled_window_add_with_viewport(
			GTK_SCROLLED_WINDOW(scrolled),
			eventbox);

	/* create message bar */
	statusbar = gtk_statusbar_new();
	gtk_box_pack_start(GTK_BOX(vbox), statusbar, FALSE, FALSE, 0);
	msg_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),"msg");
	gtk_statusbar_push(GTK_STATUSBAR(statusbar), msg_id, "-");

	gtk_signal_connect(GTK_OBJECT(vbox), "destroy",
			nc81_graph_destroy_event, NULL);

	return vbox;
}
/****************************************************************************/
GtkWidget* nc81_graph_ref_task_new(void)
{
	assert(nc81_ref_name != NULL);
	view = nc81_graph_view_ref_task;
	/* rtdl */
	return nc81_graph_new("rtl -d \"%r\"", 
			"rtdl -t \"%r\"");
}
/****************************************************************************/
GtkWidget* nc81_graph_ref_job_new(void)
{
	assert(nc81_ref_name != NULL);
	assert(nc81_task_name != NULL);
	view = nc81_graph_view_ref_job;
	/* rjdl */
	return nc81_graph_new("rjl -d \"%r\" \"%t\"", 
			"rjdl -t \"%r\" \"%t\"");
}
/****************************************************************************/
GtkWidget* nc81_graph_exp_task_new(void)
{
	assert(nc81_day_name != NULL);
	assert(nc81_task_name != NULL);
	view = nc81_graph_view_exp_task;
	/* xtdl */
	return nc81_graph_new("xtl -ds \"%d\"", 
			"xtdl -ts \"%r\"");
}
/****************************************************************************/
GtkWidget* nc81_graph_exp_job_new(void)
{
	assert(nc81_day_name != NULL);
	view = nc81_graph_view_exp_job;
	/* xjdl */
	return nc81_graph_new("xjl -ts \"%d\" \"%j\"", 
			"xjdl -ts \"%d\" \"%t\"");
}
/****************************************************************************/
